/*
 * Functions for CALL-WWL and DXCC-WWL databases
 * Imported from project tucnak (http://tucnak.nagano.cz)
 * Thanks to Ladislav Vaiz OK1ZIA!
 */

#include "osdep.h"
#ifdef HAVE_GLIB2

#include "cdwdb.h"
#include "convert.h"
#include "main.h"
#include <glib.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>


struct cw *cw;
static gchar *cw_next;
static gchar *wc_next;

struct cw *init_cw(){
    struct cw *cw;

    cw = g_new0(struct cw, 1);
    
    cw->cw = g_hash_table_new(g_str_hash, g_str_equal);
    cw->wc = g_hash_table_new(g_str_hash, g_str_equal);
    
    cw_next = NULL;
    wc_next = NULL;
    cw->latest = 0;

    return cw;
}

gboolean free_cw_item(gpointer key, gpointer value, gpointer user_data){
    struct cw_item *cwi;
    
    g_free(key);
    cwi = (struct cw_item *)value;
    if (cwi->wwl0) g_free(cwi->wwl0);
    if (cwi->wwl1) g_free(cwi->wwl1);
    g_free(value);
    return 0;    
}

gboolean free_wc_item(gpointer key, gpointer value, gpointer user_data){
    struct wc_item *wci;
    
    g_free(key);
    wci = (struct wc_item *)value;
    if (wci->call0) g_free(wci->call0);
    if (wci->call1) g_free(wci->call1);
    g_free(value);
    return 0;
        
}

void free_cw(struct cw *cw){
    g_hash_table_foreach_remove(cw->wc, free_wc_item, NULL);
    g_hash_table_destroy(cw->wc);
    
    g_hash_table_foreach_remove(cw->cw, free_cw_item, NULL);
    g_hash_table_destroy(cw->cw);
    g_free(cw);
}

void add_cw(struct cw *cw, gchar *call, gchar *wwl, gint stamp, gchar *qrv_str){
    struct cw_item *cwi;
    char *c;
    
    if (stamp<0) return; 
    if (stamp>cw->latest) cw->latest=stamp;
    
    cwi = g_hash_table_lookup(cw->cw, call);
    if (!cwi){
        cwi = g_new0(struct cw_item, 1);
        g_hash_table_insert(cw->cw, g_strdup(call), cwi);
    }
    
    if (qrv_str){ /* can be NULL*/
        for (c=qrv_str;*c!='\0';c++){
            *c=toupper(*c);
            cwi->qrv|=1<<(*c-'A');
        }
    }
    
    if (stamp < cwi->stamp1) return;
                
    if (stamp < cwi->stamp0){
        if (strcasecmp(cwi->wwl0, wwl)==0) return;

        if (cwi->wwl1) g_free(cwi->wwl1);
        cwi->wwl1   = g_strdup(wwl);
        cwi->stamp1 = stamp;
        return;
    }

    if (cwi->wwl0 && strcasecmp(cwi->wwl0, wwl)==0) {
        cwi->stamp0 = stamp;
        return;
    }

    if (cwi->wwl1) g_free(cwi->wwl1);
    cwi->wwl1   = cwi->wwl0;
    cwi->stamp1 = cwi->stamp0;
    
    cwi->wwl0   = g_strdup(wwl);
    cwi->stamp0 = stamp;
}


void add_wc(struct cw *cw, gchar *wwl, gchar *call, gint stamp){
    struct wc_item *wci;
    
    if (stamp<0) return; 
    if (stamp>cw->latest) cw->latest=stamp;
    
    wci = g_hash_table_lookup(cw->wc, wwl);
    if (!wci){
        wci = g_new0(struct wc_item, 1);
        g_hash_table_insert(cw->wc, g_strdup(wwl), wci);
    }

    if (stamp < wci->stamp1) return;

    if (stamp < wci->stamp0){
        if (strcasecmp(wci->call0, call)==0) return;

        if (wci->call1) g_free(wci->call1);
        wci->call1  = g_strdup(call);
        wci->stamp1 = stamp;
        return;
    }

    if (wci->call0 && strcasecmp(wci->call0, call)==0) {
        wci->stamp0 = stamp;
        return;
    }
    
    
    if (wci->call1) g_free(wci->call1);
    wci->call1  = wci->call0;
    wci->stamp1 = wci->stamp0;
    
    wci->call0  = g_strdup(call);
    wci->stamp0 = stamp;
}


gchar *find_wwl_by_call(struct cw *cw, gchar *call){
    struct cw_item *cwi;

    if (!call) return cw_next;
    cw_next = NULL;
        
    cwi = g_hash_table_lookup(cw->cw, call);
    if (!cwi) return NULL;
    
    cw_next = cwi->wwl1;
    return cwi->wwl0;
}

gchar *find_call_by_wwl(struct cw *cw, gchar *wwl){
    struct wc_item *wci;

    if (!wwl) return wc_next;
    wc_next = NULL;
        
    wci = g_hash_table_lookup(cw->wc, wwl);
    if (!wci) return NULL;
    
    wc_next = wci->call1;
    return wci->call0;
}

#define CW_DELIM " \t\r\n"

void load_one_cw(struct cw *cw, gchar *s){
    gchar *call, *wwl, *stamp_str, *qrv_str;
    char *token_ptr;
    
    strupr(s);
    call = strtok_r(s, CW_DELIM, &token_ptr);
    if (!call) return;

    wwl = strtok_r(NULL, CW_DELIM, &token_ptr);
    if (!wwl) return;

    stamp_str = strtok_r(NULL, CW_DELIM, &token_ptr);
    if (!stamp_str) return;
    
    qrv_str = strtok_r(NULL, CW_DELIM, &token_ptr);
    /*dbg("qrv_str(%s)=%s\n", call, qrv_str);*/

    add_cw(cw, call, wwl, atoi(stamp_str), qrv_str); /* qrv can be null */
    add_wc(cw, wwl, call, atoi(stamp_str));
}

int load_cw_from_file(struct cw *cw, gchar *filename){
    FILE *f;
    char s[102];
    int recs = 0;

    f = fopen(filename, "rt");
    if (!f){
        eprintf("! Can't open '%s'\n", filename);
        return -1;
    }

    while((fgets(s, 100, f))!=NULL){
        load_one_cw(cw,s);
        recs++;
    }

    fclose(f);
    if (verbose_mode > 0)
	    fprintf(inf, "Loaded Call-WWL records: %u\n", recs);
    return 0;
}

#endif /* HAVE_GLIB2 */
