/*
    qsomap.c - All QSO & QTH map
    Copyright (C) 2015 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "main.h"
#include "qsomap.h"

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>


struct n{
    int cnt;
};

struct cpoint{
  short int lat,lon;
  signed char c;
}__attribute__((packed));

static int minx, maxx, miny, maxy;
static int w, h;

static char *mare[] = {"JO11OT", "JO74EE", "JN65ND", "KN53KC", "KN86FC", 
    "KN40CQ", "LN51LO", "LN61TD", "LN94TU", "LN43LV", "KP50PB", "KP71RQ", 
    "KO98GK", "KP80TE", "MN65XT", "LM28PD", "JN44JI", "JO43CS", "JO22EQ",
    "KN20NU", "KN32VM", "KN10JH", "IO53JE", "IO54TM", "MN86WM", "JP53GM",
    "JP53NW", "MN86AP", "NN18NE", "MN82EJ", "MN96NQ", "KO47AW", "KO38QK",
    "KO38AH", "LN17ER", "LN18OJ", "LO45TH", "LO44HM", NULL};

static void minmax(gpointer key, gpointer value, gpointer user_data);
void putpixel(char *bmp, int x, int y, char c);
char getpixel(char *bmp, int x, int y);
void line(char *bmp, int x1, int y1, int x2, int y2, char color);
void floodfill(char *bmp, int x, int y, int c);
void priorityputpixel(char *bmp, int x, int y, char c);
void plotcor(char *bmp);

void make_qsomap(int band, FILE *htm){
    struct CALL_INFO *c;
    struct QSO_INFO *q;
    char loc[10];
    int num = 0;
    
    for (c = first_call[band]; c; c = c->next){
        if (c->categ == 0) continue;
        num++;
    }

    if (!num)
        return;


    GHashTable *calls = g_hash_table_new(g_str_hash, g_str_equal);
    GHashTable *locs = g_hash_table_new(g_str_hash, g_str_equal);

    
    int maxcnt = 0;
    for (c = first_call[band]; c != NULL; c = c->next){
        for (q = c->first_qso; q != NULL; q = q->next){
            if (strcmp(q->r_call->call, q->ok_call->call)) continue;
            if (q->r_call->loc != q->ok_call->loc) continue;
            if (!(q->r_call->loc & 0x80000000)) continue;
            if (q->wwl_outside) continue;

            char *call = q->r_call->call;
            num2loc(q->r_call->loc, loc);

            if (g_hash_table_lookup(calls, call) == NULL){
                //printf("%s %s ", call, loc);
                g_hash_table_insert(calls, call, GINT_TO_POINTER(1));

                struct n *n = g_hash_table_lookup(locs, loc);
                if (!n){
                    n = g_new0(struct n, 1);
                    g_hash_table_insert(locs, g_strdup(loc), n);
                }
                n->cnt++;
                //printf("%d\n", n->cnt);
                if (maxcnt < n->cnt) maxcnt = n->cnt;
            } 
        }
    }
//    printf("maxcnt=%d\n", maxcnt);
    minx = 10000;
    maxx = 0;
    miny = 10000;
    maxy = 0;
    g_hash_table_foreach(locs, minmax, NULL);
    minx -= 20;
    miny -= 20;
    maxx += 20;
    maxy += 20;
/*    minx = 0;
    miny = 0;
    maxx = 4319;
    maxy = 4319;*/
    //printf("x=%d..%d  y=%d..%d\n", minx, maxx, miny, maxy);

    w = maxx - minx + 1;
    h = maxy - miny + 1;
    char *bmp = g_new0(char, w * h);
    int x, y, i;
    line(bmp, 0, 0, w - 1, 0, 'w');
    line(bmp, 0, h - 1, w - 1, h - 1, 'w');
    line(bmp, 0, 0, 0, h - 1, 'w');
    line(bmp, w - 1, 0, w - 1, h - 1, 'w');
        
    // plot map
    FILE *f;
    plotcor(bmp);

    // fill mare
    char **l;
    for (l = mare; *l != NULL; l++){
        strcpy(loc, *l);
        x = ((loc[0] - 'A') * 240 + (loc[2] - '0') * 24 + (loc[4] - 'A')) - minx;
        y = maxy - ((loc[1] - 'A') * 240 + (loc[3] - '0') * 24 + (loc[5] - 'A'));
        floodfill(bmp, x, y, 'm');
        //putpixel(bmp, x, y, '9');
    }
    
    // plot QSOs
    for (c = first_call[band]; c != NULL; c = c->next){
        for (q = c->first_qso; q != NULL; q = q->next){
            if (strcmp(q->r_call->call, q->ok_call->call)) continue;
            if (q->r_call->loc != q->ok_call->loc) continue;
            if (!(q->r_call->loc & 0x80000000)) continue;
            if (q->wwl_outside) continue;

            num2loc(q->r_call->loc, loc);
            char loc2[10];
            num2loc(q->s_call->loc, loc2);
            //printf("%s %s\n", loc, loc2);
        
            x = ((loc[0] - 'A') * 240 + (loc[2] - '0') * 24 + (loc[4] - 'A')) - minx;
            y = maxy - ((loc[1] - 'A') * 240 + (loc[3] - '0') * 24 + (loc[5] - 'A'));

            int x2 = ((loc2[0] - 'A') * 240 + (loc2[2] - '0') * 24 + (loc2[4] - 'A')) - minx;
            int y2 = maxy - ((loc2[1] - 'A') * 240 + (loc2[3] - '0') * 24 + (loc2[5] - 'A'));

            line(bmp, x, y, x2, y2, 'y');
        }
    }

    plotcor(bmp);

    
    // plot locs
    for (y = maxy; y >= miny; y--){
        for (x = minx; x <= maxx; x++){
            sprintf(loc, "%c%c%c%c%c%c", 
                    'A' + x / 240,
                    'A' + y / 240,
                    '0' + (x / 24) % 10,
                    '0' + (y / 24) % 10,
                    'A' + x % 24,
                    'A' + y % 24);
            struct n *n = g_hash_table_lookup(locs, loc);
            if (n != NULL){
                int cnt = n->cnt;
                if (cnt > 9) cnt = 9;
                char oldc = getpixel(bmp, x - minx, maxy - y);
                priorityputpixel(bmp, x - minx, maxy - y, '0' + cnt);
                priorityputpixel(bmp, x - minx + 1, maxy - y, '0' + cnt);
                priorityputpixel(bmp, x - minx - 1, maxy - y, '0' + cnt);
                priorityputpixel(bmp, x - minx, maxy - y + 1, '0' + cnt);
                priorityputpixel(bmp, x - minx, maxy - y - 1, '0' + cnt);
            }
        }
    }


    // write xpm
    char *s = g_strdup_printf("%s/qsomap%d.xpm", cfg_epath, band);
    f = fopen(s, "wt");
    if (f){
        fprintf(f, "static char *image[] = {\n");
        fprintf(f, "\"%d %d %d %d\",\n", maxx - minx + 1, maxy - miny + 1, 14, 1);
        fprintf(f, "\". c #ffffff\",\n");
        fprintf(f, "\"w c #b0b0b0\",\n");
        fprintf(f, "\"m c #B0E0FF\",\n");
        fprintf(f, "\"g c #00FF00\",\n");
        fprintf(f, "\"y c #E0E0A0\",\n");
        for (i = 1; i <= 9; i++) fprintf(f, "\"%d c #%02x%02x%02x\",\n", i, 192+7*i, 0, 0);
        for (y = 0; y < h; y++){
            fprintf(f, "\"");
            for (x = 0; x < w; x++){
                int c = getpixel(bmp, x, y);
                if (c == '\0'){
                    fprintf(f, ".");
                }else{
                    fprintf(f, "%c", c);
                }
            }
            fprintf(f, "\",\n");
        }
        fprintf(f, "};\n");
        fclose(f);

    int pid = fork();
    if (pid == 0){
        //printf("child\n");
#ifdef __CYGWIN__   
        char *xpm = g_strdup_printf("%s\\qsomap%d.xpm", cfg_epath, band);
        char *png = g_strdup_printf("%s\\qsomap%d.png", cfg_epath, band);
        int ret = execl("convert.exe", "convert.exe", xpm, png, NULL);
#else
        char *xpm = g_strdup_printf("%s/qsomap%d.xpm", cfg_epath, band);
        char *png = g_strdup_printf("%s/qsomap%d.png", cfg_epath, band);
        int ret = execlp("convert", "convert", xpm, png, NULL);
#endif
        eprintf("Can't run convert\n");
        exit(0);    
    }else{
        //printf("parent: wait(%d)\n", pid);
            wait(NULL);
        //printf("child exited\n");
        //printf("unlinking '%s'\n", s);
            unlink(s);

            char text[200];
            if (band >= BAND_GHZ) {
                if (band >= BAND_GHZ_ZERO)
                    sprintf(text, "%2.0f GHz", band2freq[band]);
                else
                    sprintf(text, "%2.1f GHz", band2freq[band]);
            } else
                sprintf(text, "%2.0f MHz", band2freq[band]);

            fprintf(htm, "<h2>QSO map for %s:</h2>\n", text);
            fprintf(htm, "<img src='qsomap%d.png' alt='QSO map for %s'>\n", band, text);
            fprintf(htm, "<br>\n");
        }
    }
    g_free(s);
    //printf("DONEEEEE\n");
}

void plotcor(char *bmp){
    int x, y, oldx, oldy;

    FILE *f = fopen("tucnakcor", "rb");
    if (f != NULL){
        struct cpoint pt;
        int minlat = 32767, minlon = 32767, maxlat = -32767, maxlon = -32767;
        int oldx, oldy;
        int color = 'w';
        while(fread(&pt, 1, 5, f) == 5){
            if (pt.lon < minlon) minlon = pt.lon;
            if (pt.lat < minlat) minlat = pt.lat;
            if (pt.lon > maxlon) maxlon = pt.lon;
            if (pt.lat > maxlat) maxlat = pt.lat;

            x = ((pt.lon + 18000) * 18*10*24) / 36000 - minx;
            y = maxy - ((pt.lat + 9000) * 18*10*24) / 18000;
            if (pt.c < 0){
                oldx = x;
                oldy = y;
/*                if (pt.c == -126) // lake
                    color = '9';
                else
                    color = 'w';*/
                continue;
            }
            if (abs(x - oldx) < 100 && abs(y - oldy) < 100){  // check for back of the Earth
                line(bmp, oldx, oldy, x, y, color);
            }
            oldx = x;
            oldy = y;
        }
        fclose(f);
        //printf("lon=%d..%d  lat=%d..%d\n", minlon, maxlon, minlat, maxlat);
    }
}

void minmax(gpointer key, gpointer value, gpointer user_data){
    char *loc = (char*)key;
    struct n *n = (struct n *)value;

    int x = (loc[0] - 'A') * 240 + (loc[2] - '0') * 24 + (loc[4] - 'A');
    int y = (loc[1] - 'A') * 240 + (loc[3] - '0') * 24 + (loc[5] - 'A');

    if (maxx < x) maxx = x;
    if (minx > x) minx = x;
    if (maxy < y) maxy = y;
    if (miny > y) miny = y;

}

void putpixel(char *bmp, int x, int y, char c){
    if (x < 0 || x >= w) return;
    if (y < 0 || y >= h) return;
    bmp[y * w + x] = c;
}

char getpixel(char *bmp, int x, int y){
    if (x < 0 || x >= w) return '\0';
    if (y < 0 || y >= h) return '\0';
    return bmp[y * w + x];
}

void line(char *bmp, int x1, int y1, int x2, int y2, char color){
    int dx, dy, p;  
    int inc, tmp;

    dx=abs(x1-x2);
    dy=abs(y1-y2);
    if (dx>=dy){
        p = 2*dy-dx;
        if (x1 >= x2){   
            tmp=x1;x1=x2;x2=tmp;
            tmp=y1;y1=y2;y2=tmp;
        }
        putpixel(bmp, x1, y1, color);
        if (y2>=y1) inc=1;
        else inc=-1;
        
        while(x1 < x2){ 
            x1++;       
            if (p<0) 
                p += 2*dy;
            else{
                y1+=inc;
                p += 2*(dy-dx);
            }
            putpixel(bmp, x1, y1, color);
        }
    }else{
        p = 2*dx-dy;
        if (y1 >= y2){
            tmp=x1;x1=x2;x2=tmp;
            tmp=y1;y1=y2;y2=tmp;
        }
        putpixel(bmp, x1, y1, color);
        
        if (x2>=x1) inc=1;
        else inc=-1;
        
        while(y1 < y2){ 
            y1++;       
            if (p<0) 
                p += 2*dx;
            else{
                x1 += inc;
                p += 2*(dx-dy);
            }
            putpixel(bmp, x1,y1, color);
        }
    }
} 


void floodfill(char *bmp, int x, int y, int color){
    char c;

    if (x < 0 || x >= w) return;
    if (y < 0 || y >= h) return;

    c = getpixel(bmp, x, y);
    if (c == '\0') putpixel(bmp, x, y, color);

    if (x > 0){
        c = getpixel(bmp, x - 1, y);
        if (c == '\0') floodfill(bmp, x - 1, y, color);
    }
    
    if (x < w){
        c = getpixel(bmp, x + 1, y);
        if (c == '\0') floodfill(bmp, x + 1, y, color);
    }
    
    if (y > 0){
        c = getpixel(bmp, x, y - 1);
        if (c == '\0') floodfill(bmp, x, y - 1, color);
    }
    
    if (y < h){
        c = getpixel(bmp, x, y + 1);
        if (c == '\0') floodfill(bmp, x, y + 1, color);
    }

}

void priorityputpixel(char *bmp, int x, int y, char c){
    char oldc = getpixel(bmp, x, y);
    if (oldc >= '0' && oldc <= '9' && c < oldc){
           
    }else{
        putpixel(bmp, x, y, c);
    }
}

